@@ -1,5 +1,5 @@
 /*
- * Copyright 2019-2022 OARC, Inc.
+ * Copyright 2019-2023 OARC, Inc.
  * Copyright 2017-2018 Akamai Technologies
  * Copyright 2006-2016 Nominum, Inc.
  * All rights reserved.
@@ -82,6 +82,9 @@ static void perf__dot_connect(struct perf_net_socket* sock)
     if (!(self->ssl = SSL_new(ssl_ctx))) {
         perf_log_fatal("SSL_new(): %s", ERR_error_string(ERR_get_error(), 0));
     }
+    if (perf_net_tls_sni && !(ret = SSL_set_tlsext_host_name(self->ssl, perf_net_tls_sni))) {
+        perf_log_fatal("SSL_set_tlsext_host_name(): %s", ERR_error_string(SSL_get_error(self->ssl, ret), 0));
+    }
     if (!(ret = SSL_set_fd(self->ssl, sock->fd))) {
         perf_log_fatal("SSL_set_fd(): %s", ERR_error_string(SSL_get_error(self->ssl, ret), 0));
     }
@@ -175,8 +178,15 @@ static ssize_t perf__dot_recv(struct perf_net_socket* sock, void* buf, size_t le
             case SSL_ERROR_WANT_READ:
                 errno = EAGAIN;
                 break;
+#if OPENSSL_VERSION_NUMBER > 0x30000000L
+            case SSL_ERROR_SSL:
+                // OpenSSL 3.0+ returns this on EOF, treat everything as bad fd and reconnect
+                errno = EBADF;
+#endif
             case SSL_ERROR_SYSCALL:
                 switch (errno) {
+                case EBADF:
+                    // treat this as a retry, can happen if sendto is reconnecting
                 case ECONNREFUSED:
                 case ECONNRESET:
                 case ENOTCONN:
